#!/usr/bin/env python3
import evdev
from evdev import ecodes as ec 
#from evdev.ecodes import *

# Map Morse Code keytext to evdev Keystrokes [keycodes, modifier]
ALT = ec.KEY_LEFTALT
CTRL = ec.KEY_LEFTCTRL
SHIFT = ec.KEY_LEFTSHIFT

DOWN = 1
UP = 0

# Store modifier_state for convenient handling and state querying
modifier_state = {ALT: UP, SHIFT: UP, CTRL: UP}

def shift(key):
    'Return a keycode pair for keystrokes that require it'
    return [key, SHIFT]


#Text that is KEY_{text.upper()}  are mapped automatically
evdev_text_to_keycode_map = {
    #equal
    #slash
    'question'      :shift(ec.KEY_SLASH),
    #comma
    'period'        :ec.KEY_DOT,
    'colon'         :shift(ec.KEY_SEMICOLON),
    #apostrophe
    'underscore'    :shift(ec.KEY_MINUS),
    'parenleft'     :ec.KEY_KPLEFTPAREN,
    'parenright'    :ec.KEY_KPRIGHTPAREN,
    'numbersign'    :ec.KEY_NUMERIC_POUND,
    #minuss
    'bar'           :shift(ec.KEY_BACKSLASH),
    #backslash
    #semicolon
    #space  	
    'Return'        :ec.KEY_ENTER,
    #BackSpace
    #Delete	
    #Left
    #Right
    #Down
    #Up
    #Home
    #End
    'Escape'    :ec.KEY_ESC,
    #Insert	
    #Tab	
    'Shift_L'   :SHIFT,
    'Control_L' :CTRL,
    'Alt_L'     :ALT,
    #F1..F10   
    #F11..F24 ?? TODO These are not handled!
    'exclam'    :shift(ec.KEY_1),
    'at'        :shift(ec.KEY_2),
    'pound'     :shift(ec.KEY_3),
    # dollar
    'percent'   :shift(ec.KEY_5),
    'caret'     :shift(ec.KEY_6),
    'ampersand' :shift(ec.KEY_7), 
    'asterisk'	:ec.KEY_NUMERIC_STAR,
    #grave     
    #'greater'   :shift(ec.KEY_PERIOD), 
    #'greater'   :shift(ec.KEY_STOP), 
    'greater'   :shift(ec.KEY_DOT), 
    #'lesser'    :shift(ec.KEY_COMMA),  We used "less" instead.
    'Page_Down' :ec.KEY_PAGEDOWN,
    'Page_Up'   :ec.KEY_PAGEUP,
    'KP_Left'   :ec.KEY_LEFT,
    'KP_Right'  :ec.KEY_RIGHT,  
    'KP_Down'   :ec.KEY_DOWN,   
    'KP_Up'     :ec.KEY_UP,
    #apostrophe
    'asciitilde'    :shift(ec.KEY_GRAVE), 
    'plus'          :ec.KEY_KPPLUS,       
    'less'          :shift(ec.KEY_COMMA),       
    'braceleft'     :ec.KEY_LEFTBRACE,
    'braceright'    :ec.KEY_RIGHTBRACE, 
    'bracketleft'   :shift(ec.KEY_LEFTBRACE),
    'bracketright'  :shift(ec.KEY_RIGHTBRACE),
#Pointer_Right    
#Pointer_Down     
#Pointer_Left    
#Pointer_Up       
    'Pointer_Button1' :ec.BTN_LEFT,   
    #'Pointer_DblClick1': '', 
    #'Pointer_Hold1'     :'', 
    'Pointer_Button2' :ec.BTN_RIGHT, 
    #'Pointer_DblClick2' :'',
    #'Pointer_Hold2'     :'',
    'Pointer_Button3' :ec.BTN_MIDDLE,  
    #'Pointer_DblClick3' :'',
    #'Pointer_Hold3'     :'',     
   }

# More toggles to consider:  ec.KEY_FN, ec.KEY_INSERT
# Do not to consider: ec.NUMLOCK

def get_keystroke(text):
    'Returns a KEY_XXX keycode or a [KEY_X, KEY_LEFTSHIFT] keycode/modifier pair for the given morse text'
    print(f'evdev_keyboard get_keystoke({text}), type={type(text)}')
    assert type(text) == str, 'Invalid text type'
    keyname = f'KEY_{text.upper()}'
    try0 = evdev_text_to_keycode_map.get(text)
    if try0:
        return try0
    try1 = ec.ecodes.get(keyname)
    if try1:
        if len(text) == 1:
            # 1..9
            if text.upper() == text.lower():
                return try1
            # a..z
            if text.lower() == text:
                return try1
            # A..Z
            if text.upper() == text:
                return try1
                #return [try1, ec.KEY_LEFTSHIFT]
        print(f"{text} -> {keyname}")
        return try1
    msg = f"edev_key ERROR: {text=} not understood"
    print(msg)
    # TODO: Should we do something better if we get a wrong key?
    #logger(msg)
    return KEY_SPACE
    #return KEY_UNKNOWN


def insert_keystroke(text):
    keystroke = get_keystroke(text)
    if type(keycode) == int:
        # Handle a bare Keycode
        insert_keycode(keystroke)
    else:
        # Handle a [Keycode, Modifer] pair
        key, modifier = keystroke
        assert type(key) == int, f'Invalid evdev_keymap.get_code {keycode=}'
        assert type(modifier) == int, f'Invalid evdev_keymap.get_code {keycode=}'
        with evdev.UInput() as ui:
            ui.write(ec.EV_KEY, modifier, DOWN)
            ui.write(ec.EV_KEY, keycode, DOWN)
            ui.write(ec.EV_KEY, keycode, UP)
            ui.write(ec.EV_KEY, modifier, UP)
            ui.syn()
        modifier_state[modifier] = UP
    return


def insert_keycode(keycode):
    'Press a key down, then up, keycode only. Modifiers are as set or by PHYSICAL keypress'
    with evdev.UInput() as ui:
        ui.write(ec.EV_KEY, keycode, DOWN)
        ui.write(ec.EV_KEY, keycode, UP)
        ui.syn


def set_keycode(keycode, value):
    'Set a stateful modifier like ALT, SHIFT or CTRL keycode to 0=UP or 1=DOWN'
    assert value in [DOWN,UP], f'Invalid modifier {value=}'
    assert keycode in modifier_state, f'Invalid modifier {keycode=}'
    #if modifier_state[keycode] == value:
    #    print(f'evdev_keyboard INFO: modifer {keycode} already at {value=}')
    with evdev.UInput() as ui:
        ui.write(ec.EV_KEY, keycode, value)
        ui.syn()
    modifier_state[keycode] = value


def get_keycod_state(keycode):
    'Get the modifier state as set by set_keycode() and insert_keystroke. Does not reflect ACTUAL keyboard state!'
    return modifier_state.get(keycode, UP)
