#!/usr/bin/python3 
import time
import evdev
from evdev import UInput, ecodes as ec

DOWN, UP = 1, 0


def insert_key(key):
    with UInput() as ui:
        ui.write(ec.EV_KEY, key, DOWN)  # KEY_A down
        ui.syn()
        ui.write(ec.EV_KEY, key, UP)  # KEY_A up
        ui.syn()


def set_key(key, state):
    assert state in [DOWN, UP], f'Invalid EV_KEY state {state}'
    with UInput() as ui:
        ui.write(ec.EV_KEY, key, state)  
        ui.syn()


def play():
    print('Inserting "a" and "b" 10 times.')

    for i in range(10):
        print(i)
        insert_key(ec.KEY_A)
        time.sleep(.5)
        insert_key(ec.KEY_B)
        time.sleep(.5)

    print('Writing a hard-coded event.')
    ev = evdev.InputEvent(1334414993, 274296, ec.EV_KEY, ec.KEY_A, 1)
    with UInput() as ui:
        ui.write_event(ev)
        ui.syn()

    print('Writing a mouse event.')
    mouse = evdev.InputDevice('/dev/input/event1')
    keybd = '/dev/input/event2'
    ui = evdev.UInput.from_device(mouse, keybd, name='keyboard-mouse-device')
    x = ui.capabilities(verbose=True).keys()
    print(f'ui.capabiliites:  {x}')


if __name__ == '__main__':
    play()
