#!/usr/bin/env python3
import getpass
import grp
import logging
import os
import re
import subprocess
import sys
from urllib.parse import unquote

import evdev
from evdev import ecodes

# Local imports 
import MorseCode
#import xmouser
#import fakex
import evdev_keyboard as keyboard

# GUI Imports
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

# Global variables
UP = 0
DOWN = 1
ALT = ecodes.KEY_LEFTALT
SHIFT = ecodes.KEY_LEFTSHIFT
CTRL = ecodes.KEY_LEFTCTRL

# Character Mapping Dictionary, maps each "<key>" to an escaped keyboard code
# https://www.oreilly.com/library/view/xlib-reference-manual/9780937175262/16_appendix-h.html
char2str =     {
    "space"      :" ",
    "Return"     :"%0D", 
    "BackSpace"  :"%7F",
    "Delete"     :"%08",
    "Tab"        :"%09",
    "Escape"     :"%1B",

    "Up"         :"%1B[A",
    "Down"       :"%1B[B",
    "Left"       :"%1B[D",
    "Right"      :"%1B[C",
    "Home"       :"%1B[1~",
    "Insert"     :"%1B[2~",
    #"Delete"     :"%1b[3~",
    "End"        :"%1B[4~",
    "Page_Up"    :"%1B[5~",
    "Page_Down"  :"%1B[6~",
    "F1"         :"%1BOP",
    "F2"         :"%1BOQ",
    "F3"         :"%1BOR",
    "F4"         :"%1BOS",
    "F5"         :"%1B15~",
    "F6"         :"%1B17~",
    "F7"         :"%1B18~",
    "F8"         :"%1B19~",
    "F9"         :"%1B20~",
    "F10"        :"%1B21~",
    "F11"        :"%1B23~",
    "F12"        :"%1B24~",
 
    "equal"      :"=",
    "plus"       :"+",
    "slash"      :"/",
    "question"   :"?",
    "comma"      :",",
    "period"     :".",
    "colon"      :":",
    "quoteright" :"'",
    "quotedbl"   :"\"",
    "underscore" :"_",
    "quoteleft"  :"`",
    "parenleft"  :"(",
    "parenright" :")",
    "numbersign" :"#",
    "minus"      :"-",
    "bar"        :"|",
    "backslash"  :"\\",
    "asterisk"   :"*",
    "semicolon"  :";",
    "ampersand"  :"&",
    "asciicircum":"^", # This is not a known name...
    "percent"    :"%",
    "dollar"     :"$",
    "at"         :"@",
    "exclam"     :"!",
    "apostrophe" :"'",
    "greater"    :">",
    "less"       :"<",
    "bracketleft":"[",
    "bracketright":"[",
    "asciitilde" :"~",

    "<none>"     :""
}

# Gdk key mappings
keyval2char =   {
    Gdk.KEY_BackSpace   :"BackSpace",
    Gdk.KEY_Tab         :"Tab",
    #Gdk.KEY_Linefeed    :"Return",
    Gdk.KEY_Return      :"Return",
    Gdk.KEY_Escape      :"Escape",
    Gdk.KEY_Delete      :"Delete",

    Gdk.KEY_Up          :"Up",
    Gdk.KEY_Down        :"Down",
    Gdk.KEY_Left        :"Left",
    Gdk.KEY_Right       :"Right",
    Gdk.KEY_Home        :"Home",
    Gdk.KEY_Insert      :"Insert",
    Gdk.KEY_End         :"End",
    Gdk.KEY_Page_Up     :"Page_Up",
    Gdk.KEY_Page_Down   :"Page_Down",

    Gdk.KEY_F1          :"F1",
    Gdk.KEY_F2          :"F2",
    Gdk.KEY_F3          :"F3",
    Gdk.KEY_F4          :"F4",
    Gdk.KEY_F5          :"F5",
    Gdk.KEY_F6          :"F6",
    Gdk.KEY_F7          :"F7",
    Gdk.KEY_F8          :"F8",
    Gdk.KEY_F9          :"F9",
    Gdk.KEY_F10         :"F10",
    Gdk.KEY_F11         :"F11",
    Gdk.KEY_F12         :"F12",

    Gdk.KEY_Shift_L     :"Shift_L",
    Gdk.KEY_Shift_R     :"Shift_L",
    Gdk.KEY_Control_L   :"Control_L",
    Gdk.KEY_Control_R   :"Control_L",
    Gdk.KEY_Alt_L       :"Alt_L",
    Gdk.KEY_Alt_R       :"Alt_L",    
    Gdk.KEY_space       :"space",
    Gdk.KEY_numbersign  :"numbersign",
    Gdk.KEY_backslash   :"backslash"
}

'''
class Myevdev:
    'To see codes, run python3 -m evdev.evtest'

    def __init__(self, stateful=None):
        self.stateful = stateful


    def insert_key(self, keycode):
        'Press a key down, then release it'
        with evdev.UInput() as ui:
            ui.write(ecodes.EV_KEY, keycode, DOWN) 
            ui.syn()
            ui.write(ecodes.EV_KEY, keycode, UP) 
            ui.syn


    def set_key(self, keycode, value):
        'Set a stateful modifier like ALT, SHIFT or CTRL keycode to 0=UP or 1=DOWN'
        assert value in [DOWN,UP], 'Invalid toggle value'
        assert not self.stateful or keycode in self.stateful, 'keycode {keycode} is not stateful'
        with evdev.UInput() as ui:
            ui.write(ecodes.EV_KEY, keycode, value) 
            ui.syn()
'''

def set_keycode(keycode, value):
    'Set persistent-state modifier keys like CTRL, ALT, SHIFT'
    return keyboard.set_keycode(keycode, value)
 

class Key_Mouse():
    enable = False

    def __init__(self, mouse_callback):
        # Control the mouse
        '''
        if False:
            self.xmouser = xmouser.XMouser(callback = mouse_callback)
            self.xthread = fakex.FakeXThread()
            self.xthread.set_mouse_callback(mouse_callback)
            self.xthread.start()
        '''

        # Load key maps
        #char2keycode = {v: k for k, v in keycode2char.items()}
        #char2keysym = {v: k for k, v in keysym2char.items()}
        self.char2keycode = {v: k for k, v in keyval2char.items()}
        self.char2keysym = {v: k for k, v in char2str.items()}
        self.check()


    def check(self):
        "Check if we are in the input group, Enable if we are!"
        retval = subprocess.run('groups', capture_output=True)
        #print(f'actual_ok = "input" in {retval.stdout.split()}')
        actual_ok = 'input' in retval.stdout.decode().split()
        user = getpass.getuser()
        retval = grp.getgrnam('input')
        set_ok = user in retval.gr_mem
        if actual_ok:
            print(f'User "{user}" is a member of the "input" group.')
            self.enable = True
        elif set_ok:
            print(f'User "{user}" is not a member of "input" until reset. Please Reboot.')
        else:
            print(f'WARNING: User "{user}" is NOT a member of the "input" group.')
            print('To enable key entry, run the following and reboot:')
            print('  sudo usermod -a -G input $USER')


    def grab_set(self, grab_state):
        "Grab events from the whole screen"
        # We now use a grab slaver...
        #self.xthread.grab_set(grab_state)
        #self.xmouser.set_grab(grab_state)
        '''
        win = self.win.get_toplevel()
        device = Gdk.Display.get_default().get_default_seat().get_pointer()
        if self.tog_grab.get():
            result = device.grab(
                win, 
                Gdk.GrabOwnership.NONE,
                False,
                Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK | Gdk.EventMask.POINTER_MOTION_MASK,
                None,
                Gdk.CURRENT_TIME
            )
            msg = f"device.grab returned {result}"
            logging.debug(msg)
            print(msg)
        else:
            device.ungrab(Gdk.CURRENT_TIME)
        '''
        ''' This code worked in Gnome 3 or earlier but not in Gnome 4.
        if self.tog_grab.get():
            mask = Gdk.ModifierType.PRESS_MASK | Gdk.ModifierType.RELEASE_MASK
            # Grab the pointer,capture button presses
            # Constrain the mouse to be inside win
            # Gdk.pointer_grab(win,owner_events, confine_to, cursor, ...)
            #   owner_events = False
            #   confine_to = None
            #   cursor = None
            # We confine to the grid because the Vte widget has its own use for the button1
            #x = Gdk.pointer_grab(win, owner_events=False, event_mask=mask, confine_to=self.grid)
            x = Gdk.device_grab(win, owner_events=False, event_mask=mask, confine_to=self.grid)
            logging.debug("pointer_grab returned %s" % x)
            self.term.set_mouse_autohide(1)
            self.grab_state = 1
            # disable the X screensaver, which morseall can't stop,
            x = os.system("xset s off")
            if x: 
                logging.debug("Turning off the screensaver returned %s" % x)
        else:
            #Gdk.pointer_ungrab(0)
            # Gdk.pointer_ugrab() Deprecated since: 3.0. Use Device_ungrab
            # Use gdk_device_ungrab(), together with gdk_device_grab() instead.
            #Gdk.device_ungrab(0)
            # Gdk.device_ungrab() Depracated since: 3.2 Use gdk_seat_ungrab()
            #Gdk.device_ungrab(0)
            Gdk.seat_ungrab(0)
            ###2010### self.term.show_pointer()
            self.grab_state = 0
        '''

    def insert_keystroke(self, text):
        keystroke = keyboard.get_keystroke(text)
        keyboard.insert_keystroke(keystroke)
        '''
        if type(keystroke) == int:
            # Handle a bare Keycode
            keyboard.insert_keycode(keystroke)
        else:
            # Handle a [Keycode, Modifer] pair
            key, modifier = keystroke
            assert type(key) == int, f'Invalid evdev_keymap.get_code {keycode=}'
            assert type(modifier) == int, f'Invalid evdev_keymap.get_code {keycode=}'
            self.keyboard.set_key(modifier, DOWN)
            self.keyboard.insert_key(key)
            self.keyboard.set_key(modifier, UP)
        return 
        '''
        '''
        keycode = self.char2keycode.get(text)
        keysym = self.char2keysym.get(text)
        kx = self.xthread.get_key_by_name(keysym or text) 
        st = unquote(char2str.get(text)) if char2str.get(text) else ''
        print(f'Insert a fake keypress "{text}":  sym={keysym} co={keycode} {kx=} {st=}')
        return [keycode, keysym, kx, st]
        if not self.enable:
            return

        if st:
            self.keyboard.insert_key(st)
        elif kx:
             self.keyboard.insert_key(kx.keycode)
        #    self.keyboard.insert_key(kx)
        #    #self.xthread.fake_keyname(text)
        elif keysym:
            self.keyboard.insert_key(keysym)
            #self.xthread.fake_keysym(keysym)
        elif keycode:
            self.keyboard.insert_key(keycode)
            #self.xthread.fake_keycode_press(keycode)
            #self.xthread.fake_keycode_release(keycode)
        else:
            print(f'Insert ERROR: unkown how to handle "{text}"')
        return
        '''


if __name__ == "__main__":
    print('test keys or something')
    set_keycode(SHIFT, UP)
